/*******************************************************************
* (c) Copyright 2000-12, Celstream Technologies Pvt Ltd.
* This program, or portions thereof, is protected as an
* unpublished work under the copyright laws of the India.
* TranslationDM.c
*
*  Created on: Dec 27, 2011
*      Author: sanjeevk
*/

#include "TranslationDM.h"
#ifdef LIBXML_TREE_ENABLED

int main()
{
    LIBXML_TEST_VERSION
    xmlDoc *doc = NULL;
    xmlNode *root_element = NULL;

    /*global structures*/
    memset(&Incs,NULL_CHARACTER,sizeof(struct Import_include));
    memset(&st_El,NULL_CHARACTER,sizeof(struct elementlist));
    memset(&rule,NULL_CHARACTER,sizeof(struct Translation_rules));
    memset(&obj,NULL_CHARACTER,sizeof(struct objectnodes));

    if(!Load_Translation_rule())
    {
    	//getch();
    	return 0;
    }

    /*PASS 1*/
    /*get the root element of schema from the file specified*/
    doc = xmlReadFile(IN_FILE, NULL, 0);
	if (doc == NULL)
	{
		printf("\nerror: could not parse file %s",IN_FILE);
		//getch();
		return 0;
	}
	root_element = xmlDocGetRootElement(doc);

	/*Use root of schema to load all the includes*/
	Incs.SR[Incs.included++]=root_element;
	printf("\nschemaLocation (includes and imports): Reading files");
	Load_all_roots(Incs.SR[0]);

	/*copy the root name to the st_El.parentname*/
	strcpy(st_El.parentname,ROOT);
	stPrint_element(ROOT,NULL,NULL);
	/*Find element*/
	printf("\nListing all elements and types.");
	Find_the_element_and_print(Incs.SR[0],ROOT);


	/* print the list of element in a file <for reference only>*/
	int i;
	FILE *fg = fopen(LIST_FILENAME,"w");
	for(i=1;i<=st_El.count;i++)
	{
		fprintf(fg,"Element:%d %s\n",i,st_El.name[i]);
	}

	/*PASS 2*/
	printf("\nCreating DM \"%s\".",DM_FILENAME);
	Trans_DM();

	fclose(fg);
	xmlFreeDoc(doc);
    xmlCleanupParser();
    xmlMemoryDump();
    return 0;
}
#else
int main(void) {
    fprintf(stderr, "Tree support not compiled in\n");
    exit(1);
}
#endif




/*searches for the element in the tree and expand its type*/
int Find_the_element_and_print(xmlNode *root_element,char *elementname)
{
    xmlNode *sch_node = NULL;
    int efound;

    for (sch_node = root_element; sch_node; sch_node = sch_node->next)
    {
    	if(compare(sch_node->name,"element"))
    	{
    		if(Is_prop_val_exist(sch_node,"name",elementname))
    		{
    			Detect_the_type_and_print(sch_node);
    			return 1;
			}
    	}
    	efound=Find_the_element_and_print(sch_node->children,elementname);
    	if(efound)
    	{
    		return efound;
    	}
    }
    return 0;
}

/*Takes the element and reads the attribute to find the name of the type and searches it and further searches in its child*/
void Detect_the_type_and_print(xmlNode * element_node)
{
	xmlNode *el_child_node;
	char typevalue[STRL];

	if(Is_prop_exist(element_node,"type",typevalue))
	{
		Namespace_remove(typevalue);
		if(XSD_types(typevalue))
		{
			stPrint_type("type",typevalue);
		}
		else
		{
			All_File_find_type_node(typevalue,"ATTRIBUTE");
		}
		return;
	}

	for (el_child_node = element_node->children; el_child_node; el_child_node = el_child_node->next)
	{
		if(compare(el_child_node->name,"simpleType")||compare(el_child_node->name,"complexType"))
		{
			Expand_type(el_child_node);
			return;
		}
	}
	return;
}

/*searches in all the trees to find the type*/
void All_File_find_type_node(char *typename,char *message)//depends on structure Incs
{
	xmlNode* type=NULL;
	int i;

	for(i=0;i<=Incs.included;i++)
	{
		type=File_find_type_node(Incs.SR[i],typename);
		if(type)
		{
			Expand_type(type);
			break;
		}

	}
	if(i>Incs.included)
	{
		printf("\nError:NOT FOUND(%s) TYPE type=\"%s\"",message,typename);
	}
	return;
}

/*searches for the type name in tree */
xmlNode * File_find_type_node(xmlNode *root_node,char *typepropname)
{
    xmlNode *sch_node = NULL,*breaker=NULL;

    for (sch_node = root_node; sch_node; sch_node = sch_node->next)
    {
    	if(compare(sch_node->name,"simpleType")||compare(sch_node->name,"complexType")||compare(sch_node->name,"group"))
    	{
    		if(Is_prop_val_exist(sch_node,"name",typepropname))
    		{
				return sch_node;
    		}
    	}
    	breaker= File_find_type_node(sch_node->children,typepropname);
    	if(breaker)
    	{
    		return breaker;
    	}
    }
    return NULL;
}




/* directs the type nodes to their handlers*/
void Expand_type(xmlNode * type_node)
{
	xmlNode *tp_node = NULL,*tp_child_node;

	tp_node=type_node;

	if(compare(tp_node->name,"simpleType"))
	{
		Handle_SimpleType(tp_node);
		return;
	}
	else
	if(compare(tp_node->name,"complexType"))
	{
		for(tp_child_node=tp_node->children;tp_child_node;tp_child_node=tp_child_node->next)
		{
			if(compare(tp_child_node->name,"choice"))
			{
				Handle_choice(tp_child_node);
			}
			if(compare(tp_child_node->name,"sequence"))
			{
				Handle_cC_sequence(tp_child_node);
			}
			if(compare(tp_child_node->name,"complexContent"))
			{
				Handle_complexContent(tp_child_node);
			}
			if(compare(tp_child_node->name,"simpleContent"))
			{
				Handle_simplecontent(tp_child_node);
			}

		}
	}
	return;
}


/* prints the simple type definition*/
void Handle_SimpleType(xmlNode *simpletypenode)
{
	xmlNode *child_node=NULL,*res_child_node=NULL;
	char basevalue[STRL];
	int j;

	for(child_node=simpletypenode->children;child_node;child_node=child_node->next)
	{
		if(compare(child_node->name,"restriction"))
		{
			if(Is_prop_exist(child_node,"base",basevalue))
			{
				Namespace_remove(basevalue);
				if(XSD_types(basevalue))
				{
					int lengthalone=0,otheralone=0;

					stPrint_type("type",basevalue);
					for(res_child_node=child_node->children;res_child_node;res_child_node=res_child_node->next)
					{
						if(compare(res_child_node->name,"enumeration"))
						{
							char enumer[STRL];
							Is_prop_exist(res_child_node,"value",enumer);
							stPrint_type("enumeration",enumer);
							otheralone++;
						}
						if(compare(res_child_node->name,"maxLength"))
						{
							char smaxlength[STRL];
							Is_prop_exist(res_child_node,"value",smaxlength);
							stPrint_type("maxLength",smaxlength);
							lengthalone++;
						}
						if(compare(res_child_node->name,"whiteSpace"))
						{
							char s[STRL];
							Is_prop_exist(res_child_node,"value",s);
							stPrint_type("whiteSpace",s);
							otheralone++;
						}
						if(compare(res_child_node->name,"minLength"))
						{
							char s[STRL];
							Is_prop_exist(res_child_node,"value",s);
							stPrint_type("minLength",s);
							lengthalone++;
						}
						if(compare(res_child_node->name,"pattern"))
						{
							char s[STRL];
							Is_prop_exist(res_child_node,"value",s);
							stPrint_type("pattern",s);
							otheralone++;
						}
					}
				}

				else
				{
					All_File_find_type_node(basevalue,"");
				}
			}

			return;
		}

		if(compare(child_node->name,"union"))
		{
			char membertypevalue[STRL]={0};

			stPrint_type("type","UNION");
			if(Is_prop_exist(child_node,"memberTypes",membertypevalue))
			{
				int i;
				char mem_type_name[20][STRL]={{0,0}};

				/*split the members*/
				char *pIn=membertypevalue;
				strcat(membertypevalue," ");
				for(i=0;pIn[0]!=NULL_CHARACTER;i++)
				{
					pIn=Split_Copy(pIn,mem_type_name[i]," ");
				}

				/*find each member*/
				for(i=0;mem_type_name[i][0];i++)
				{
					Namespace_remove(mem_type_name[i]);
					if(XSD_types(mem_type_name[i]))
					{
						stPrint_type("type",mem_type_name[i]);
					}
					All_File_find_type_node(mem_type_name[i],"ST_union member");
				}
			}
			return;
		}
	}
	return;
}




/*prints the simple content extension base type and restriction base type*/
void Handle_simplecontent(xmlNode *simplecontent_node)
{
	xmlNode *child_node=NULL,*res_child_node=NULL;
	char basevalue[STRL];

	for(child_node=simplecontent_node->children;child_node;child_node=child_node->next)
	{
		if(compare(child_node->name,"extension"))
		{
			if(Is_prop_exist(child_node,"base",basevalue))
			{
				Namespace_remove(basevalue);
				if(XSD_types(basevalue))
				{
					stPrint_type("type",basevalue);
				}
				else
				{
					All_File_find_type_node(basevalue,"CT_SC EXTENSION BASE");
				}
			}

			for(res_child_node=child_node->children;res_child_node;res_child_node=res_child_node->next)
			{
				if(compare(res_child_node->name,"simpleType"))
				{
					Handle_SimpleType(res_child_node);
				}
			}
		return;
		}

		if(compare(child_node->name,"restriction"))
		{
			if(Is_prop_exist(child_node,"base",basevalue))
			{
				Namespace_remove(basevalue);
				if(XSD_types(basevalue))
				{
					stPrint_type("type",basevalue);
				}
				else
				{
					All_File_find_type_node(basevalue,"CT_SC RESTRICTION BASE");
				}
			}
			for(res_child_node=child_node->children;res_child_node;res_child_node=res_child_node->next)
			{
				if(compare(res_child_node->name,"simpleType"))
				{
					Handle_SimpleType(res_child_node);
				}
			}

			return;
		}
	}
	return;
}

/*prints the complex content extension base type and the sequence in its child*/
void Handle_complexContent(xmlNode *complexContent_node)
{
	xmlNode *child_node=NULL,*ex_child_node=NULL;
	char basevalue[STRL];

	for(child_node=complexContent_node->children;child_node;child_node=child_node->next)
	{
		if(compare(child_node->name,"extension"))
		{
			if(Is_prop_exist(child_node,"base",basevalue))
			{
				Namespace_remove(basevalue);
				All_File_find_type_node(basevalue,"CT_CC EXTENSION BASE");
			}

			for(ex_child_node=child_node->children;ex_child_node;ex_child_node=ex_child_node->next)
			{
				if(compare(ex_child_node->name,"sequence"))
				{
					Handle_cC_sequence(ex_child_node);
				}
			}/*expand all the elements in the sequence of extension*/
		}
	}/*search for extension in the complex type*/
	return;
}


/*directs the childs of sequence node to element printer and sequence and choice handlers*/
void Handle_cC_sequence(xmlNode *sequence_node)
{
	xmlNode *child_node=NULL;

	for(child_node=sequence_node->children;child_node;child_node=child_node->next)
	{
		if(compare(child_node->name,"element"))
		{
			Element_name_printer(child_node);

		}
		if(compare(child_node->name,"sequence"))
		{
			Handle_cC_sequence(child_node);
		}

		if(compare(child_node->name,"choice"))
		{
			Handle_choice(child_node);
		}
	}/*for all the nodes in the sequence*/
	return;
}



/*directs the childs of choice node to element printer and sequence and choice handlers*/
void Handle_choice(xmlNode *choice_node)
{
	xmlNode *child_node=NULL;

	for(child_node=choice_node->children;child_node;child_node=child_node->next)
	{
		if(compare(child_node->name,"element"))
		{
			Element_name_printer(child_node);
		}
		if(compare(child_node->name,"sequence"))
		{
			Handle_cC_sequence(child_node);
		}
		if(compare(child_node->name,"choice"))
		{
			Handle_choice(child_node);
		}
		if(compare(child_node->name,"group"))
		{
			Handle_group(child_node);
		}
	}
}

/* handles element groups called only from choice-special case*/
void Handle_group(xmlNode *group_ref_node)
{
	xmlNode *child_node=NULL,*group_node=NULL;
	char refvalue[STRL];

	if(Is_prop_exist(group_ref_node,"ref",refvalue))
	{
		Namespace_remove(refvalue);
		int i;
		for(i=0;i<=Incs.included;i++)
		{
			if(File_find_type_node(Incs.SR[i],refvalue))
			{
				group_node=File_find_type_node(Incs.SR[i],refvalue);
				break;
			}

		}
	}

	for(child_node=group_node->children;child_node;child_node=child_node->next)
	{
		if(compare(child_node->name,"element"))
		{
			Element_name_printer(child_node);
		}
		if(compare(child_node->name,"sequence"))
		{
			Handle_cC_sequence(child_node);
		}
	}
	return;
}


/*prints the name of the element and searches in all the files if it is defined there itself it detects the type*/
void Element_name_printer(xmlNode *el_node)//depends on structure st_El, Incs
{
	char temp[STRL],refnamevalue[STRL];

	if(Is_prop_exist(el_node,"ref",refnamevalue)||Is_prop_exist(el_node,"name",refnamevalue))
	{
		Namespace_remove(refnamevalue);
		strcpy(temp,st_El.parentname);

		if(Is_prop_val_exist(el_node,"maxOccurs","unbounded"))
		{
			snprintf(st_El.parentname,STRL,"%s.%s.{i}",st_El.parentname,Namingoptimize(st_El.parentname,refnamevalue));
		}
		else
		{
			snprintf(st_El.parentname,STRL,"%s.%s",st_El.parentname,Namingoptimize(st_El.parentname,refnamevalue));
		}

		char m[STRL]={0},n[STRL]={0};
		Is_prop_exist(el_node,"maxOccurs",m);
		Is_prop_exist(el_node,"minOccurs",n);

		stPrint_element(st_El.parentname,m,n);



		if(Is_prop_exist(el_node,"name",refnamevalue))
		{
			Detect_the_type_and_print(el_node);
		}
		else
		{
			int i;
			for(i=0;i<=Incs.included;i++)
			{
				if(Find_the_element_and_print(Incs.SR[i],refnamevalue))
				{
					break;//only the first element definition is taken
				}
			}
			if(i>Incs.included)
			{
				printf("\nError:NOT FOUND ELEMENT \"%s\"",refnamevalue);
			}
		}

		strcpy(st_El.parentname,temp);
	}
	return;
}


/* finds the object name from the parameter name and adds the parameter to the object*/
void CWMP_find_obj_add_param(char *fullpname,xmlNodePtr parameternode)//depends on structure obj
{
	int n;

	for(n=strlen(fullpname);n>=0;n--)
	{
		if(fullpname[n]=='.' && fullpname[n+1]!='{')
		{
			fullpname[n]=0;break;
		}
	}

	for(n=0;n<obj.count;n++)
	{
		if(compare(fullpname,obj.objectnodename[n]))
		{
			xmlAddChild(obj.objectnode[n], parameternode);
			break;
		}
	}

	return;
}

/* reads all the type properties from the element list file an creates the syntax node*/
void CWMP_param_add_syntax(int el_num,xmlNodePtr parameternode,int list)//depends on structure st_El
{
	int i,IsTypeAdded=0,IsmaxLengthAdded=0,IsUnion=0,enum_alone=1;
	xmlNodePtr syntax_node = NULL,type_node=NULL,list_node=NULL;

	syntax_node = xmlNewChild(parameternode, NULL, BAD_CAST "syntax", BAD_CAST "");
	if(list)
	{
		list_node = xmlNewChild(syntax_node, NULL, BAD_CAST "list", BAD_CAST "");
	}

	for(i=0;strlen(st_El.typeprop[el_num][i]);i++)
	{
		xmlNodePtr restr_node;

		if(strstr(st_El.typeprop[el_num][i],"type"))
		{
			char typevalue[STRL];
			strcpy(typevalue,st_El.typevalue[el_num][i]);
			if(compare(typevalue,"UNION"))
			{
				IsUnion=1;
				//Merge union
				{
					char set[10][500][STRL];
					memset(set,NULL_CHARACTER,sizeof(set));
					int setl[10]={0};
					int j,setcount=-1,internal=0;
					for(j=1;strlen(st_El.typeprop[el_num][j]);j++)
					{
						if(strstr(st_El.typevalue[el_num][j],"UNION"))
						{
							continue;
						}
						if(strstr(st_El.typeprop[el_num][j],"type") )
						{
							setcount++;
							internal=0;
						}
						setl[setcount]=internal+1;
						strcpy(set[setcount][internal++],st_El.typeprop[el_num][j]);
					}

					for(j=0;j<=setcount;j++)
					{
						int m,enumpresent=0;
						for(m=0;m<setl[j];m++)
						{
							if(strstr(set[j][m],"enumeration"))
							{
								enumpresent=1;
								break;//when the member has enum leave the member
							}
						}
						if(!enumpresent)
						{
							enum_alone=0;
							break;//when any one member has no enum
						}
					}
				}//merge union
			}
			else
			{
				if(!IsTypeAdded)
				{
					if(!IsUnion)
					{
						int j;
						enum_alone=0;
						for(j=0;strlen(st_El.typeprop[el_num][j]);j++)
						{
							if(strstr(st_El.typeprop[el_num][j],"enumeration"))
							{
								enum_alone=1;
								break;
							}
						}
					}
					type_node=xmlNewChild(syntax_node, NULL, BAD_CAST XSD_types(typevalue), BAD_CAST "");
					IsTypeAdded=1;
				}
			}
		}
		if(strstr(st_El.typeprop[el_num][i],"enumeration") && enum_alone)
		{
			restr_node=xmlNewChild(type_node, NULL, BAD_CAST "enumeration", BAD_CAST "");
			xmlNewProp(restr_node, BAD_CAST "value", BAD_CAST st_El.typevalue[el_num][i]);
		}
		if(strstr(st_El.typeprop[el_num][i],"maxLength") && !IsmaxLengthAdded && !enum_alone)
		{
			restr_node=xmlNewChild(type_node, NULL, BAD_CAST "size", BAD_CAST "");
			xmlNewProp(restr_node, BAD_CAST "maxLength", BAD_CAST st_El.typevalue[el_num][i]);
			IsmaxLengthAdded=1;
		}
	}
}


/* reads the element list file and creates the CWMP heirarchy*/
void Trans_DM()//depends on structure st_El,obj
{
	xmlDocPtr doc = NULL;/* document pointer */
	xmlNodePtr model_node = NULL, parameternode = NULL,documentnode=NULL;/* node pointers */
	char buff[STRL],read[STRL];
	int list=0,el_num=0;

	obj.count=0;

	doc = xmlNewDoc(BAD_CAST "1.0");
	documentnode= xmlNewNode(NULL, BAD_CAST "document");
	xmlDocSetRootElement(doc, documentnode);

	model_node=xmlNewChild(documentnode, NULL, BAD_CAST "model", BAD_CAST "");
	xmlNewProp(model_node, BAD_CAST "isService", BAD_CAST "true");
	xmlNewProp(model_node, BAD_CAST "name", BAD_CAST ROOT);

	xmlNewChild(model_node, NULL, BAD_CAST "description", BAD_CAST "Write a description");

	el_num=1;
	while(el_num<=st_El.count)
	{
		strcpy(read,st_El.name[el_num]);
		if(strlen(read)>1)
		{

			if(!strlen(st_El.typeprop[el_num][0]))
			{
				obj.objectnode[obj.count] = xmlNewChild(model_node, NULL, BAD_CAST "object", BAD_CAST "");
				strcpy(obj.objectnodename[obj.count],read);
				sprintf(buff,"%s.",read);
				xmlNewProp(obj.objectnode[obj.count], BAD_CAST "name", BAD_CAST buff);

					if(strlen(st_El.minoccur[el_num]))
						xmlNewProp(obj.objectnode[obj.count], BAD_CAST "minEntries", BAD_CAST st_El.minoccur[el_num]);
					else
						xmlNewProp(obj.objectnode[obj.count], BAD_CAST "minEntries", BAD_CAST "1");

					if(strlen(st_El.maxoccur[el_num]))
					{
						xmlNewProp(obj.objectnode[obj.count], BAD_CAST "maxEntries", BAD_CAST st_El.maxoccur[el_num]);
						xmlNewProp(obj.objectnode[obj.count], BAD_CAST "numEntriesParamter", BAD_CAST "NumberOfEntries");
					}
					else
						xmlNewProp(obj.objectnode[obj.count], BAD_CAST "maxEntries", BAD_CAST "1");

				sprintf(buff, "Write a description about object %s",read);
				xmlNewChild(obj.objectnode[obj.count], NULL, BAD_CAST "description", BAD_CAST buff);

				obj.count++;
			}
			else
			{
				char pname[STRL],pfullname[STRL];
				int n;

				strcpy(pfullname,read);

				for(n=strlen(pfullname);n>=0;n--)
				{
					if(pfullname[n]=='.' && pfullname[n+1]!='{')
					{
						strcpy(pname,&pfullname[n+1]);
						break;
					}
				}

				parameternode = xmlNewNode(NULL, BAD_CAST "parameter");
				if(strstr(pname,".{"))
				{
					strcpy(strstr(pname,".{"),"");
					list=1;
				}
				xmlNewProp(parameternode, BAD_CAST "name", BAD_CAST pname);
				xmlNewProp(parameternode, BAD_CAST "access", BAD_CAST "readonly");
				xmlNewProp(parameternode, BAD_CAST "activeNotify", BAD_CAST "forceEnabled");

				sprintf(buff, "Write a description about parameter %s",pname);
				xmlNewChild(parameternode, NULL, BAD_CAST "description", BAD_CAST buff);

				CWMP_param_add_syntax(el_num,parameternode,list);
				list=0;

				CWMP_find_obj_add_param(pfullname, parameternode);

			}
		}
		el_num++;
	}
	 xmlSaveFormatFileEnc(DM_FILENAME, doc, "UTF-8", 1);
	 return;
}
